import logging

from django.core.cache import cache
from django.utils import timezone

from autojob.models import JobList

logger = logging.getLogger(__name__)


def job_before(func):
    def wrapper(*args, **kwargs):
        from autojob.job_new import JobAction
        job_state = JobList.objects.filter(id=int(args[1])).values('job_state')[0]['job_state']
        if job_state == 1:
            job_value = cache.get('modify_' + args[0])
            if job_value:
                JobAction.modify_job(args[0], job_value)
                cache.delete('modify_' + args[0])
            if get_lock(args[0]):
                if args[2] == 'date' or (job_value and job_value[0] == 'date'):
                    JobList.objects.filter(id=int(args[1])).update(job_state=0, update_time=timezone.now())
                func(*args, **kwargs)
            else:
                JobAction.stop_job(args[0])
        else:
            JobAction.stop_job(args[0])

    return wrapper


# Job status detection
def monitor_job_status():
    from autojob.job_new import JobAction
    job_stop = cache.iter_keys('stop_*')
    for i in job_stop:
        JobAction.stop_job(cache.get(i))
    job_modify = cache.iter_keys('modify_*')
    for i in job_modify:
        JobAction.modify_job(i.replace('modify_', ''), cache.get(i))


def get_lock(lock_name):
    lock = cache.lock('lock.' + lock_name, timeout=5)
    return lock.acquire(blocking=False)


def job_control():
    global logger
    try:
        from autojob.job_new import JobAction
        from autojob import job_new
        from autojob.models import JobList
        import logging

        logger = logging.getLogger(__name__)

        job_list_ = JobList.objects.filter(job_state=1)
        logger.info('Program startup, start scheduled tasks : ' + str(job_list_.__len__()))
        job_new.start_scheduler()
        for job_ in job_list_:
            if hasattr(JobAction, 'start_%s_job' % job_.action_type):
                func = getattr(JobAction(), 'start_%s_job' % job_.action_type)
                func(job_.trigger.trigger_func, job_.job_rate, str(job_.id))
            else:
                logger.error('【start error】 ' + job_.trigger.trigger_func + ' does not exists!')
    except Exception as rel:
        logger.error("scheduler start error:" + str(rel))



# coding=utf-8
import datetime
import logging

from autojob.job_tool import job_before

logger = logging.getLogger(__name__)


@job_before
def test_job(*args):
    # This is the task you need to perform
    now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    logger.info(now + '__This is a timed task for testing:' + args[0])



# coding=utf-8
import logging
import os

from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.events import EVENT_JOB_EXECUTED, EVENT_JOB_ERROR, EVENT_JOB_MISSED

from autojob.models import JobList

logging.getLogger('apscheduler.executors.default').setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


def start_scheduler():
    global scheduler
    scheduler = BackgroundScheduler()
    scheduler.start()
    from .job_tool import monitor_job_status
    scheduler.add_job(monitor_job_status, 'cron', second=0, minute='*/1', hour='*', day='*', month='*', day_of_week='*',
                      year='*', id='monitor_job_status-0', replace_existing=True)


class JobAction:
    def __init__(self):
        self.scheduler = scheduler
        self.scheduler.add_listener(self.my_listener, EVENT_JOB_ERROR | EVENT_JOB_MISSED | EVENT_JOB_EXECUTED)

    # start job for date
    @staticmethod
    def start_date_job(trigger, job_rate, id):
        exec(JobList.objects.get(id=id).trigger.func_path)
        trigger_id = trigger + '-' + id
        scheduler.add_job(eval(trigger), 'date', run_date=job_rate, id=trigger_id, args=[trigger_id, id, 'date'],
                          replace_existing=True, coalesce=True)

    # start job for cron
    @staticmethod
    def start_cron_job(trigger, job_rate, id):
        exec(JobList.objects.get(id=id).trigger.func_path)
        rate = job_rate.split()
        trigger_id = trigger + '-' + id
        # second minute hour day month week year (The parameters must be in this order)
        scheduler.add_job(eval(trigger), 'cron', second=rate[0], minute=rate[1], hour=rate[2], day=rate[3],
                          month=rate[4], day_of_week=rate[5], year=rate[6], id=trigger_id,
                          args=[trigger_id, id, 'cron'], replace_existing=True, coalesce=True)

    # stop job
    @staticmethod
    def stop_job(trigger_id):
        if scheduler.get_job(trigger_id):
            scheduler.remove_job(trigger_id)

    # pause job
    @staticmethod
    def pause_job(trigger_id):
        logger.info(trigger_id)
        scheduler.pause_job(trigger_id)

    # restart job
    @staticmethod
    def resume_job(trigger_id):
        scheduler.resume_job(trigger_id)

    # modify job
    @staticmethod
    def modify_job(trigger_id, job_value):
        if scheduler.get_job(trigger_id):
            logger.info('modify job：' + trigger_id + '\t' + str(job_value))
            if job_value[0] == 'cron':
                rate = job_value[1].split()
                scheduler.reschedule_job(trigger_id, trigger='cron', second=rate[0], minute=rate[1], hour=rate[2],
                                         day=rate[3], month=rate[4], day_of_week=rate[5], year=rate[6])
            elif job_value[0] == 'date':
                scheduler.reschedule_job(trigger_id, trigger='date', run_date=job_value[1])

    @staticmethod
    def get_jobs():
        jobs = [item.id for item in scheduler.get_jobs()]
        logger.info(str(os.getpid()) + ' - all jobs：' + str(jobs))

    @staticmethod
    def my_listener(event):  # Add listener
        job = scheduler.get_job(event.job_id)
        if not event.exception:
            pass
        else:
            logger.error("jobname=%s|jobtrigger=%s|errcode=%s|exception=[%s]|traceback=[%s]|scheduled_time=%s",
                         job.name, job.trigger, event.code, event.exception, event.traceback, event.scheduled_run_time)
        # scheduler.shutdown(wait=False)



from apscheduler.jobstores.base import ConflictingIdError, JobLookupError
from django.contrib import admin
from django.core.cache import cache
from django.utils import timezone

from autojob.job_new import JobAction

from .models import JobList, JobListTrigger
import logging
from django.contrib import messages

logger = logging.getLogger(__name__)


# Register your models here.
@admin.register(JobListTrigger)
class JobListTriggerAdmin(admin.ModelAdmin):
    list_display = ('id', 'trigger_name', 'trigger_func', 'func_path', 'description')
    list_display_links = ('id', 'trigger_name')
    list_per_page = 20
    list_max_show_all = 7
    ordering = ('id',)


# Register your models here.
@admin.register(JobList)
class JobListAdmin(admin.ModelAdmin):
    list_display = ('id', 'job_name', 'job_state', 'action_type', 'job_rate', 'update_time')
    list_display_links = ('id', 'job_name')
    list_filter = ('job_state',)
    list_editable = ('action_type', 'job_rate',)
    list_per_page = 20
    list_max_show_all = 7
    ordering = ('-job_state', 'id',)
    actions = ['start_job', 'stop_job']

    def get_readonly_fields(self, request, obj=None):
        if obj:
            readonly_fields = ('trigger_id', 'job_state', 'update_time',)
        else:
            readonly_fields = ('job_state',)
        return readonly_fields

    def save_model(self, request, obj, form, change):
        if obj.job_state == 1:
            cache.set('modify_' + obj.trigger.trigger_func + '-' + str(obj.id), (obj.action_type, obj.job_rate),
                      timeout=60)
        obj.save()

    def start_job(self, request, queryset):
        global job_name
        trigger_list = queryset.values('trigger_id')
        rate_list = queryset.values('job_rate')
        id_list = queryset.values('id')
        job_num = len(id_list)
        name_list = queryset.values('job_name')
        action_type = queryset.values('action_type')
        update_time = queryset.values('update_time')
        current_time = timezone.now()
        for (trigger, rate, id_, name, type_, time_) in zip(trigger_list, rate_list, id_list, name_list, action_type,
                                                            update_time):
            try:
                job_name = name.get('job_name')
                job_rate = rate.get('job_rate')
                job_id = id_.get('id')
                trigger_name = JobList.objects.get(id=job_id).trigger.trigger_func
                job_type = type_.get('action_type')
                last_time = time_.get('update_time')
                interval_time = current_time - last_time
                seconds = interval_time.days * 24 * 3600 + interval_time.seconds
                if seconds < 60:
                    messages.warning(request, 'Please start %s again with %ss' % (job_name, str(60 - seconds)))
                    job_num -= 1
                    continue
                if hasattr(JobAction(), 'start_%s_job' % job_type):
                    func = getattr(JobAction(), 'start_%s_job' % job_type)
                    func(trigger_name, job_rate, str(job_id))
                    queryset.filter(id=job_id).update(job_state=1, update_time=current_time)
                else:
                    messages.error(request, '【start failed】: ' + job_name + ' job does not exists!!')
                    job_num -= 1
            except ConflictingIdError as e:
                logger.error(e)
                messages.error(request, job_name + ' job startup conflict occurred!!')
                job_num -= 1
            except Exception as e:
                logger.error(e)
                messages.error(request, job_name + ' job start error！<br/>Check the scheduling type or frequency')
                job_num -= 1
        JobAction().get_jobs()
        if job_num > 0 and len(id_list) > 1:
            messages.success(request,
                             ('the rest' if job_num != len(
                                 id_list) else 'selected') + ' %s jobs were successfully started' % str(job_num))
        elif job_num == len(id_list) == 1:
            messages.success(request, '%s job started successfully' % job_name)
        return messages

    start_job.short_description = '启动任务'
    start_job.icon = 'el-icon-video-play'
    start_job.type = 'success'

    def stop_job(self, request, queryset):
        global job_name
        name_list = queryset.values('job_name')
        trigger_list = queryset.values('trigger')
        id_list = queryset.values('id')
        job_num = len(id_list)
        current_time = timezone.now()
        for (trigger, id_, name) in zip(trigger_list, id_list, name_list):
            try:
                job_name = name.get('job_name')
                job_id = id_.get('id')
                trigger_name = JobList.objects.get(id=job_id).trigger.trigger_func
                JobAction.stop_job(trigger_name + '-' + str(job_id))
                queryset.filter(id=job_id).update(job_state=0, update_time=current_time)
                cache.set('stop_' + str(job_id), trigger_name + '-' + str(job_id), timeout=60)
            except JobLookupError as e:
                logger.error(e)
                job_num -= 1
                messages.error(request, job_name + " job does not exists!！！")
            except Exception as e:
                logger.error(e)
                job_num -= 1
                messages.error(request, job_name + ' job stop error！')
        JobAction().get_jobs()
        if job_num > 0 and len(id_list) > 1:
            messages.success(request,
                             ('the rest' if job_num != len(
                                 id_list) else 'selected') + ' %s jobs were successfully stopped' % str(job_num))
        elif job_num == len(id_list) == 1:
            messages.success(request, "%s has been stopped" % job_name)
        return messages

    stop_job.short_description = '停止任务'
    stop_job.icon = 'el-icon-video-pause'
    stop_job.type = 'danger'



import os

from setuptools import setup, find_packages

# with open("README.md") as f:
#     long_description = f.read()

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="autojob",
    version='0.5.1',
    description="Django customizes management pages based on APScheduler",
    keywords='autojob',
    # long_description=README,
    long_description_content_type="text/markdown",
    author='zkrliumy',
    author_email='zkrliumy@163.com',
    url="https://github.com/lmycc/autojob",
    include_package_data=True,
    # 你要安装的包，通过 setuptools.find_packages 找到当前目录下有哪些包
    packages=find_packages(include=('autojob',)),
    zip_safe=False,
    # 表明当前模块依赖哪些包，若环境中没有，则会从pypi中下载安装
    install_requires=[
        'Django>=2.0',
        'djangorestframework>=3.7.0',
        'APScheduler',
        'django-redis',
    ],

    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Framework :: Django",
        "Framework :: Django :: 2",
        "Framework :: Django :: 3",
        "Framework :: Django :: 4",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



from django.db import models
from django.utils import timezone


class JobListTrigger(models.Model):
    trigger_name = models.CharField('trigger_name', max_length=25)
    trigger_func = models.CharField('trigger_func', max_length=25)
    func_desc = 'e.p.：from autojob.test_job import test_job'
    func_path = models.CharField('func_path', help_text=func_desc, max_length=255)
    description = models.CharField('description', max_length=50)

    def __str__(self):
        return self.trigger_name

    class Meta:
        verbose_name = '定时任务触发器'
        verbose_name_plural = verbose_name


class JobList(models.Model):
    job_name = models.CharField('job_name', max_length=25)
    trigger = models.ForeignKey(JobListTrigger, on_delete=models.SET_NULL, null=True)
    type_choices = (('date', 'date'), ('cron', 'cron'))
    type_content = '''Parameter corresponding to scheduling type (execution frequency) ：<br/>
        1、date：Execute task at 1am on August 30, 2019<br/>value：2019-8-30 01:00:00 <br/>
        2、cron：Execute tasks at 2:00 am every day(second minute hour day month week year (The parameters must 
        be in this order))<br/>value：0 0 2 * * * *'''
    action_type = models.CharField('scheduling type', choices=type_choices, max_length=25, default='cron')
    gender_choices = ((0, 'stop'), (1, 'start'),)
    job_state = models.IntegerField('job status', choices=gender_choices, default=0)
    job_rate = models.CharField('frequency', help_text=type_content, max_length=50)
    time = models.IntegerField('number of times', default=0)
    help_text_update_time = '''The task status change before the time is updated<br/>
                    start job: The current time must be greater than one minute from the last time it was changed<br/>
                    stop job: No need to wait to restart'''
    update_time = models.DateTimeField('update time', auto_now_add=timezone.now(), help_text=help_text_update_time)

    def __str__(self):
        return self.job_name

    class Meta:
        verbose_name_plural = '定时任务'
